/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/mman.h>

#include "libppm.h"

static char *getchunk(FILE *f);
static int ppm_read_header(FILE *infp, struct ppm_image *p);

int ppminit(struct ppm_image *p) {
  if (!p) return -1;
  p->data = NULL;
  p->width = p->height = p->greyscale = p->mapped = 0;
  return 0;
}

int ppmalloc(struct ppm_image *p, int width, int height, int greyscale) {
  if (!p) return -1;
  if (p->data) free(p->data);
  if ((p->data = (PEL*)malloc(width*height*(greyscale ? 1 : 3))) == NULL) {
    p->width = p->height = 0;
    return -1;
  }
  p->width = width;
  p->height = height;
  p->greyscale = greyscale;
  p->mapped = 0;
  return 0;
}
  
int ppmdestroy(struct ppm_image *p) {
  if (!p || !p->data) return -1;
  if (p->mapped) {
    if (munmap(p->data, p->width * p->height * (p->greyscale ? 1 : 3)) == -1)
      perror("munmap");
    p->mapped = 0;
  }
  else
    free(p->data);
  p->data = NULL;
  return 0;
}

int ppmread(FILE *infp, struct ppm_image *p) {
  int n, bytes, bytesread = 0;

  if (!infp || !p) return -1;
  if (p->data) ppmdestroy(p);

  if (ppm_read_header(infp, p) == -1) {
    fprintf(stderr, "Input does not appear to be in PPM format\n");
    return -1;
  }

  bytes = p->width * p->height * (p->greyscale ? 1 : 3);
  p->data = malloc(bytes);
  while (bytesread < bytes) {
    n = fread(p->data + bytesread, 1, bytes - bytesread, infp);
    if (n <= 0) return -1;
    bytesread += n;
  }
  p->mapped = 0;

  return 0;
}

int ppmmap(FILE *infp, struct ppm_image *p) {
  int bytes;

  if (!infp || !p) return -1;
  if (p->data) ppmdestroy(p);

  if (ppm_read_header(infp, p) == -1) {
    fprintf(stderr, "Input does not appear to be in PPM format\n");
    return -1;
  }

  bytes = p->width * p->height * (p->greyscale ? 1 : 3);
  p->data = mmap(NULL, bytes, PROT_READ, MAP_SHARED, fileno(infp), 0);
  if ((int)p->data == -1) { perror("mmap"); return -1; }
  p->data += ftell(infp);
  p->mapped = 1;

  return 0;
}

int ppmwrite(FILE *outfp, struct ppm_image *img) {
  if (!outfp || !img || !img->data || img->width <= 0 || img->height <= 0)
    return -1;

#if 0
  /* check automatically to see if the image is greyscale */
  greyscale = 1;
  for (i=0; i<img->width*img->height; i++)
    if (img->data[i*3] != img->data[i*3+1]
        || img->data[i*3] != img->data[i*3+2]) {
      greyscale = 0;
      break;
    }
#endif
  
  if (img->greyscale) {
    fprintf(outfp, "P5\n%d %d 255\n", img->width, img->height);
    if (fwrite(img->data, img->width*img->height, 1, outfp) <= 0) return -1;
  }
  else {
    fprintf(outfp, "P6\n%d %d 255\n", img->width, img->height);
    if (fwrite(img->data, img->width*img->height*3, 1, outfp) <= 0) return -1;
  }
  return 0;
}

void ppmcopy(struct ppm_image *to, struct ppm_image *from) {
	ppmalloc(to, from->width, from->height, from->greyscale);
	memcpy(to->data, from->data,
		to->width * to->height * (to->greyscale ? 1 : 3));
}

int ppmscale(struct ppm_image *p, int nw, int nh) {
  PEL *nimg;
  int x, y, c;
	int bypp;  /* bytes per pixel */

	bypp = p->greyscale ? 1 : 3;

  if (!p || !p->data || nw <= 0 || nh <= 0) return -1;

  nimg = (unsigned char *)malloc(nw * nh * bypp);
  if (!nimg) return -1;

  for (y=0; y<nh; y++)
    for (x=0; x<nw; x++) {
      int tmp1, tmp2;
      tmp1 = (x + y*nw) * bypp;
      tmp2 = (x*p->width/nw + y*p->height/nh*p->width) * bypp;
			for (c=0; c<bypp; c++)
				nimg[tmp1+c] = p->data[tmp2+c];
    }

  ppmdestroy(p);
  p->width = nw;
  p->height = nh;
  p->data = nimg;
  return 0;
}

int ppmcrop(struct ppm_image *p, int nw, int nh, PEL bg_r, PEL bg_g, PEL bg_b) {
  PEL *nimg, *exp;
  int x, y;
  int uh, uw;

  if (p->greyscale) {
    fprintf(stderr, "Cropping of greyscale images not supported.\n");
    return -1;
  }

  if (!p || !p->data || nw <= 0 || nh <= 0) return -1;

  nimg = (unsigned char *)malloc(nw * nh * 3);
  if (!nimg) return -1;
  uh = (nh < p->height) ? nh : p->height;
  uw = (nw < p->width) ?  nw : p->width;

  for (y=0; y<uh; y++) {
    memcpy(&nimg[y*nw*3], &p->data[y*p->width*3], uw*3);
    exp = &nimg[y*nw*3+uw*3];
    for (x=0; x<nw-uw; x++) {
      *exp++ = bg_r;
      *exp++ = bg_g;
      *exp++ = bg_b;
    }
  }
  exp = &nimg[uh*nw*3];
  for (x=0; x<(nh-uh)*nw; x++) {
    *exp++ = bg_r;
    *exp++ = bg_g;
    *exp++ = bg_b;
  }

  ppmdestroy(p);
  p->width = nw;
  p->height = nh;
  p->data = nimg;
  return 0;
}

int ppmgreyscale(struct ppm_image *p) {
  int i;
  if (!p || !p->data) return -1;
  if (p->greyscale) return 0;
  if (p->mapped) {
    fprintf(stderr, "Cannot modify mapped images.\n");
    return -1;
  }
  for (i=0; i<p->width*p->height; i++)
    p->data[i] = (p->data[i*3] + p->data[i*3+1] + p->data[i*3+2])/3;
  p->data = realloc(p->data, p->width * p->height);
  p->greyscale = 1;
  return 0;
}

int ppmdither2(struct ppm_image *p) {
  int i, dither_right=0, *dither_down;
  PEL *q = p->data;
  if (!p || !p->data) return -1;

  if (p->greyscale) {
    fprintf(stderr, "Dithering of greyscale images not supported.\n");
    return -1;
  }

  dither_down = (int*)malloc(sizeof(int)*p->width);
  if (!dither_down) return -1;
  bzero(dither_down, sizeof(int)*p->width);

  for (i=0; i<p->width*p->height; i++,q+=3) {
    int x = i%p->width, sum;
    if (!x) dither_right=0;
    sum = dither_right + dither_down[x] + *q + *(q+1) + *(q+2);
    if (sum <= 381) {
      *q = *(q+1) = *(q+2) = 0;
      dither_right = dither_down[x] = sum/2;
    }
    else {
      *q = *(q+1) = *(q+2) = 255;
      dither_right = dither_down[x] = (sum-765)/2;
    }
  }
  free(dither_down);
  return 0;
}

static char *getchunk(FILE *f) {
  static char buf[200] = "\0";
  static char *p = buf;
  char *q, *tmp;
  if (!*p) {
    do
      fgets(buf, 200, f);
    while (*buf == '#');
    p = buf;
  }
  while (*p == ' ' || *p == '\t') p++;
  q = p;
  while (*q != ' ' && *q != '\t' && *q != '\n' && *q != '\0') q++;
  tmp = p;
  if (*q)     
    p = q+1;
  else
    p = q;
  *q = '\0';
  return tmp;
}

static int ppm_read_header(FILE *infp, struct ppm_image *p) {
  char *buf;
  int ign;

  buf = getchunk(infp);
  if (!strcmp("P5", buf))
    p->greyscale = 1;
  else if (!strcmp("P6", buf))
    p->greyscale = 0;
  else
    return -1;
  buf = getchunk(infp);   p->width  = atoi(buf);
  buf = getchunk(infp);   p->height = atoi(buf);
  buf = getchunk(infp);   ign       = atoi(buf);
  if (ign != 255)
    return -1;

  return 0;
}
