/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

#ifndef LIBPPM_H
#define LIBPPM_H

#include <stdio.h>

typedef unsigned char PEL;
#define PPM_PIXEL(img, x, y, c) (img).data[3*((x) + (y)*(img).width) + c]
#define PPM_GPIXEL(img, x, y) (img).data[(x) + (y)*(img).width]

struct ppm_image {
  int width, height;
	unsigned int greyscale:1, mapped:1;
  PEL *data;
};

/* initialize a PPM structure with an empty picture */
int ppminit(struct ppm_image *p);

/* allocate a picture buffer for an initialized PPM structure */
int ppmalloc(struct ppm_image *p, int width, int height, int greyscale);

/* deallocate a picture buffer */
int ppmdestroy(struct ppm_image *p);

/* read a PPM image from a file pointer into an initialized PPM structure */
/* automatically allocates a picture buffer */
/* uses 8-bit greyscale or 24-bit RGB PGM/PPM "raw bits" format only */
int ppmread(FILE *infp, struct ppm_image *p);

/* read a PPM image from a file pointer into an initialized PPM structure */
/* maps the image buffer to the disk file instead of reading it into */
/* allocated memory. */
/* uses 8-bit greyscale or 24-bit RGB PGM/PPM "raw bits" format only */
int ppmmap(FILE *infp, struct ppm_image *p);

/* write the PPM image to a file pointer */
/* uses 24-bit RGB PPM "raw bits" format only */
int ppmwrite(FILE *outfp, struct ppm_image *img);

/* copy one PPM image into another */
void ppmcopy(struct ppm_image *to, struct ppm_image *from);

/* scale a PPM image to a the new width and height given by (nw, nh) */
/* automatically deallocate the old picture buffer and allocate a new one */
/* scaling uses simple pixel duplication/removal -- no anti-aliasing or */
/* interpolation */
int ppmscale(struct ppm_image *p, int nw, int nh);

/* crop/grow a PPM image to the new width and height given by (nw, nh) */
/* automatically deallocate the old picture buffer and allocate a new one */
/* if the new image is larger, fill in empty pixels with (bg_r, bg_g, bg_b) */
/* if the new image is smaller, throw out pixels */
int ppmcrop(struct ppm_image *p, int nw, int nh, PEL bg_r, PEL bg_g, PEL bg_b);

/* switch a PPM image from RGB to greyscale */
/* still uses 24-bit "raw bits" encoding format */
/* intensity = (red + green + blue) / 3 */
int ppmgreyscale(struct ppm_image *p);

/* dither a PPM image to black (#000000) and white (#ffffff) */
/* still uses 24-bit "raw bits" encoding format */
/* uses a simplified version of Floyd-Steinberg: round each value and */
/* spill error onto next pixel below and next pixel to the right */
int ppmdither2(struct ppm_image *p);

#endif
