/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>
#include "libppm.h"
#include "pgpf.h"

extern void ppmdither(struct ppm_image *ppm);

void ppm2pgpf(const struct ppm_image *ppm, Image **image, Frame **frame,
		Tile ***tile, int *ntiles, int id, int tilesize, int compress,
		int indexed);
static Tile *make_tile(const struct ppm_image *ppm, int x, int y,
		int tile_size, int compress, int indexed);
static int write_header(FILE *ofp, const char *filename);
static int write_record_headers(FILE *ofp, const Image *image,
		const Frame *frame, Tile **tile, int ntiles, int pos);
static void write_inf(FILE *ofp, const Image *image);
static void write_fri(FILE *ofp, const Frame *frame);
static void write_til(FILE *ofp, const Tile *tile);
static GString *rle_compress(GString *str, int chunk_size);
static void usage(const char *prog);

int main(int argc, char **argv) {
	struct ppm_image ppm, thumb_ppm;
	FILE *fp;
	Image *image, *t_image;
	Frame *frame, *t_frame;
	Tile **tile, **t_tile;
	int ntiles, t_ntiles, pos, i, compress=1, tilesize=64, indexed=FALSE;
	char c;

	while ((c = getopt(argc, argv, "cit:")) != -1)
		switch (c) {
			case 'c':
				compress = 0;
				break;
			case 'i':
				indexed = TRUE;
				break;
			case 't':
				tilesize = atoi(optarg);
				g_assert(tilesize >= 8 && tilesize <= 128);
				break;
			default:
				usage(argv[0]);
		}
	if (optind != argc-1)
		usage(argv[0]);

	ppminit(&ppm);
	if (isatty(0)) {
		if ((fp = fopen(argv[optind], "r")) == NULL) {
			perror(argv[optind]);
			exit(1);
		}
		ppmread(fp, &ppm);
		fclose(fp);
	}
	else
		ppmread(stdin, &ppm);
	ppminit(&thumb_ppm);
	ppmcopy(&thumb_ppm, &ppm);
	if (ppm.width > ppm.height)
		ppmscale(&thumb_ppm, 44, 44*ppm.height/ppm.width);
	else
		ppmscale(&thumb_ppm, 44*ppm.width/ppm.height, 44);
	ppmcrop(&thumb_ppm, 44, 44, 0, 0, 0);

	if (indexed) ppmdither(&ppm);
	ppmdither(&thumb_ppm);

	ppm2pgpf(&thumb_ppm, &t_image, &t_frame, &t_tile, &t_ntiles, IMG_THUMB,
		tilesize, compress, TRUE);
	ppm2pgpf(&ppm, &image, &frame, &tile, &ntiles, IMG_MAIN, tilesize, compress,
		indexed);

	pos = write_header(stdout, argv[optind]);
	writeint(stdout, ntiles + t_ntiles + 4, 2);  pos+=2;  /* number of records */

	pos += 10*(ntiles+t_ntiles+4);
	pos += write_record_headers(stdout, t_image, t_frame, t_tile, t_ntiles, pos);
	pos += write_record_headers(stdout, image, frame, tile, ntiles, pos);
	write_inf(stdout, t_image);
	write_fri(stdout, t_frame);
	for (i=0; i<t_ntiles; i++)
		write_til(stdout, t_tile[i]);
	write_inf(stdout, image);
	write_fri(stdout, frame);
	for (i=0; i<ntiles; i++)
		write_til(stdout, tile[i]);

	return 0;
}

void ppm2pgpf(const struct ppm_image *ppm, Image **image, Frame **frame,
		Tile ***tile, int *ntiles, int id, int _tilesize, int compress,
		int indexed) {
	int ntilesx, ntilesy, x, y, tileid, tilesize;

	*image = g_new(Image, 1);
	(*image)->id = id;
	(*image)->width = ppm->width;
	(*image)->height = ppm->height;
	(*image)->bpp = indexed ? 8 : 16;
	(*image)->tilesize = tilesize = 
		MAX(MIN(ppm->height, _tilesize), MIN(ppm->width, _tilesize));
	(*image)->compr = compress ? COMPR_RLE : COMPR_NONE;
	(*image)->frameid = id;
	(*image)->paletteid = 0;

	*frame = g_new(Frame, 1);
	(*frame)->id = id;
	(*frame)->nframes = 1;
	(*frame)->tileid = g_new(int, 1);
	(*frame)->tileid[0] = id;

	ntilesx = (ppm->width+(tilesize-1))/tilesize;
	ntilesy = (ppm->height+(tilesize-1))/tilesize;
	*ntiles = ntilesx * ntilesy;
	*tile = g_new(Tile*, *ntiles);
	tileid = id;
	for (y=0; y<ntilesy; y++)
		for (x=0; x<ntilesx; x++) {
			(*tile)[y*ntilesx + x] =
				make_tile(ppm, x*tilesize, y*tilesize, tilesize, compress, indexed);
			(*tile)[y*ntilesx + x]->id = tileid++;
		}
}

static Tile *make_tile(const struct ppm_image *ppm, int left, int top,
		int tile_size, int compress, int indexed) {
	int x, y;
	Tile *ret = g_new(Tile, 1);
	GString *data = g_string_new("");
	static const unsigned char black[] = { 0, 0, 0 };

	for (y=top; y<top+tile_size; y++)
		for (x=left; x<left+tile_size; x++)
			if (indexed) {
				if (y<ppm->height && x<ppm->width)
					g_string_append_c(data, PPM_GPIXEL(*ppm, x, y));
				else
					g_string_append_c(data, 230 /* black */);
			}
			else {
				const unsigned char *p;
				unsigned short pixel;
				if (y<ppm->height && x<ppm->width)
					p = &PPM_PIXEL(*ppm, x, y, 0);
				else
					p = black;
				pixel = ((*p & 0xF8) << 8)
							+ ((*(p+1) & 0xFC) << 3)
							+ ((*(p+2) & 0xF8) >> 3);
				g_string_append_c(data, (pixel>>8));
				g_string_append_c(data, (pixel&0xFF));
			}

	if (compress) {
		GString *rle_data = rle_compress(data, indexed ? 1 : 2);
		g_string_free(data, TRUE);
		ret->data = rle_data->str;
		ret->length = rle_data->len;
		g_string_free(rle_data, FALSE);
	}
	else {
		ret->data = data->str;
		ret->length = data->len;
		g_string_free(data, FALSE);
	}
	return ret;
}

static int write_header(FILE *ofp, const char *filename) {
	char name[33];
	const char *basename;
	const char *header =
		"\x00\x01"          /* attributes */
		"\x02\x00"          /* version */
		"\xbb\x10\x72\x76"  /* creation date */
		"\xbb\x10\x72\x78"  /* modification date */
		"\xbb\x10\x72\x78"  /* backup date */
		"\x00\x00\x00\x56"  /* modification number */
		"\x00\x00\x00\x00"  /* app info */
		"\x00\x00\x00\x00"  /* sort info */
		"IMVSIMVS"          /* database type and creator */
		"\x00\x00\x00\x00"  /* unique ID */
		"\x00\x00\x00\x00"; /* next record list */

	memset(name, 0, 33);
	if ((basename = strrchr(filename, '/')) != NULL)
		basename++;
	else
		basename = filename;
	if (!strcasecmp(basename+strlen(basename)-4, ".ppm"))
		strncpy(name, basename, MIN(strlen(basename)-4, 32));
	else
		strncpy(name, basename, MIN(strlen(basename), 32));

	fwrite(name, 1, 32, ofp);
	fwrite(header, 1, 44, ofp);

	return 32 + 44;
}

static int write_record_headers(FILE *ofp, const Image *image,
		const Frame *frame, Tile **tile, int ntiles, int pos) {
	int i;
	int oldpos = pos;

	fprintf(ofp, "iINF");
	writeint(ofp, image->id, 2);
	writeint(ofp, pos, 4);
	pos += 18;

	fprintf(ofp, "iFRI");
	writeint(ofp, frame->id, 2);
	writeint(ofp, pos, 4);
	pos += 6 + 2*frame->nframes;

	for (i=0; i<ntiles; i++) {
		fprintf(ofp, "iTIL");
		writeint(ofp, tile[i]->id, 2);
		writeint(ofp, pos, 4);
		pos += 2 + tile[i]->length;
	}

	return pos - oldpos;
}

static void write_inf(FILE *ofp, const Image *image) {
	/* image info (iINF) header */
	writeint(ofp, image->width, 2);
	writeint(ofp, image->height, 2);
	writeint(ofp, 0, 2);                    /* dummy1 */
	writeint(ofp, image->bpp, 2);
	writeint(ofp, 0, 2);                    /* dummy2 */
	writeint(ofp, image->tilesize, 2);
	writeint(ofp, image->compr, 2);
	writeint(ofp, image->frameid, 2);
	writeint(ofp, image->paletteid, 2);
}

static void write_fri(FILE *ofp, const Frame *frame) {
	int i;

	/* frame info (iFRI) header */
	writeint(ofp, frame->nframes, 2);
	writeint(ofp, 0, 2);                  /* frame-to-frame wait time */
	writeint(ofp, 0, 2);                  /* flags */
	for (i=0; i<frame->nframes; i++)
		writeint(ofp, frame->tileid[i], 2);
}

static void write_til(FILE *ofp, const Tile *tile) {
	writeint(ofp, tile->length, 2);
	fwrite(tile->data, 1, tile->length, ofp);
}

static GString *rle_compress(GString *str, int chunk_size) {
	GString *data = g_string_new("");
	int pos, peekpos, i;

	for (pos=0; pos<str->len; pos=peekpos) {
		for (peekpos=pos+chunk_size; peekpos<str->len && peekpos<pos+257*chunk_size
				&& !memcmp(str->str+pos, str->str+peekpos, chunk_size); peekpos+=chunk_size) ;

		for (i=0; i<chunk_size; i++)
			g_string_append_c(data, str->str[pos+i]);
		if (peekpos > pos+chunk_size) {
			for (i=0; i<chunk_size; i++)
				g_string_append_c(data, str->str[pos+i]);
			g_string_append_c(data, (peekpos-pos)/chunk_size-2);
		}
	}
	
	return data;
}

static void usage(const char *prog) {
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "  %s [options] foo.ppm > foo.prc\n", prog);
	fprintf(stderr, "  %s [options] 'Image name' < foo.ppm > foo.prc\n\n", prog);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -c        Do not compress the image\n");
	fprintf(stderr, "  -i        Generate a dithered 8-bpp image\n");
	fprintf(stderr, "  -t<size>  Specify tile size\n");

	exit(1);
}
